# -*- coding: utf-8 -*-
#
# Copyright 2017-2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset providers."""

import abc
from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:
    from renku.core.dataset.providers.api import ProviderApi


class IDatasetProviderPlugin(abc.ABC):
    """Abstract class for a dataset provider plugin."""

    @classmethod
    @abc.abstractmethod
    def dataset_provider(cls) -> "Type[ProviderApi]":
        """Supported dataset provider plugin.

        Returns:
            a ProviderApi class definition (not instance) from a specific plugin
        """
        raise NotImplementedError
